#!/bin/env python3
import os
import sys
import textwrap
import argparse
import getpass
import netrc
import datetime
from os.path import basename
import urllib
from urllib.parse import urlencode
from urllib.request import urlopen, urlretrieve
from urllib import request
from http.cookiejar import CookieJar
import requests
import json
import math
import yaml

### The lines below are to get the IP address.
### We may want to make this static and assign a fixed value to the IPAddr variable
import socket
hostname = socket.gethostname()
IPAddr   = socket.gethostbyname(hostname)
######################################


# Dictionary with PODAAC data set IDs
CCIDDICT                                 = {}

CCIDDICT["SMM_MOG"]                      = "SWOT_XDF_MOG2D_COMBINEE_1.0" #
CCIDDICT["SMM_PO1"]                      = "SWOT_XDF_HISTO_POLE_1.0" # PODAAC does not have correct PO1 file currently

# Dictionary trying to provide user-friendly descriptions of data sets
CCIDDESCRIPTIONS                         = {}
CCIDDESCRIPTIONS["SMM_MOG"]              = "SWOT MOG 2-D (DAC) Auxiliary File"
CCIDDESCRIPTIONS["SMM_PO1"]              = "SWOT PO1 Pole Location Auxiliary File (old format)" # Need to find PO1 file


AVAILABLE_CCIDS = list(CCIDDICT.keys())

CMR      = "cmr.earthdata.nasa.gov"
MACHINE  = "urs.earthdata.nasa.gov"
PAGE_MAX = 2000
MINMAX   = 999999999

UPDATEFILE        = ".update"
DEFAULT_SLEEP_MIN = 60
DEFAULT_TSPAN_MIN = 60
TOKEN_URL         = "https://"+CMR+"/legacy-services/rest/tokens"
USER_NAME, _, PASSWORD = netrc.netrc().authenticators(MACHINE)

###############################################################################
# FUNCTION: parseArgs
###############################################################################
def parseArgs():
    """
    This function parses the user input arguments to run getSWOTData.py

    Help details using:'python getSWOTData.py'

    Function returns the parsed command line arguments.
    """

    # Define the description of the script
    description = textwrap.dedent("""
    This script can be used to pull SWOT data from PODAAC.
    Use -h for help.""")

    # Define command-line arguments
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument(
        '-p', metavar='prodid', dest='prodid', type=str, required=False,
        help="ID of product to download. Can be the PODAAC ccid if not one of \
        the following options: %s."%", ".join(AVAILABLE_CCIDS),
        default=None)

    parser.add_argument(
        '-pdesc', dest='prodiddesc', required=False, action='store_true',
        help="Provide list of products IDs and descriptions for download. If not on list used PODAAC ccid.")

    parser.add_argument(
        '-listonly', dest='listonly', required=False, action='store_true',
        help="Provide list of available products within specified window and with provided products. Do not download files.")

    parser.add_argument(
        '-d', metavar=('download_dir'), dest='ddir', type=str, required=False,
        help=('Directory to download data.'))

    parser.add_argument(
        '-t', metavar=('token'), dest='token', type=str, required=False,
        help=('Token string for data access to PODAAC. If not provided it will\
        be created at each instance.'),
        default=None,)

    parser.add_argument(
        '-s', metavar=('sleep'), dest='sleep', type=int,
        help=('Time window to sleep between check for new data download in minutes.\
        Default is %s minutes. If value of 0 is provided then perform one time\
        download and exit.' % (DEFAULT_SLEEP_MIN)),
        default=DEFAULT_SLEEP_MIN,)

    parser.add_argument(
        '-e', metavar='file_extensions', dest='extlist', type=str, nargs='+',
        help=('List of file extensions of files to download. Default downloads \
        all files with the specified product ID. Could use .nc to download netCDF\
        files with .nc extension only, ISPData.dat to download raw telemetry files\
        only, .rnx.gz to download GNSS RINEX files only, .rnx to download GNSS-RO \
        RINEX files, .EOF to download orbit solution or orbit rev files.'),
        default=[],)

    parser.add_argument(
        '-f', metavar='file_patterns', dest='pattlist', type=str, nargs='+',
        help=('List of patterns within names of files to download. Default \
        downloads all files with the specified product ID. Could use S6A to \
        download files with S6A pattern in file name only.'),
        default=[],)

    parser.add_argument(
        '-w', metavar='twin', dest='twin', type=int,
        help=('Time window from now to look back for new data in minutes.\
        Default is %s minutes. Only used for first download. Subsequent\
        downloads pull data since the prior download time which is in a\
        file named %s in the download directory' % (DEFAULT_TSPAN_MIN, UPDATEFILE)),
        default=DEFAULT_TSPAN_MIN)

    parser.add_argument(
        "-c", "--cycle", required=False, dest="search_cycles", nargs="*",
        help="Cycle number for determining downloads.", type=int)

    parser.add_argument(
        "--pass", required=False, dest="search_pass",
        help="Pass number for determining downloads.", type=int)

    parser.add_argument(
        '-l', metavar='logfile', dest='logfile', type=str,
        help=('Name of logfile for download history. Default is sys.__stdout__\
        writing directly to terminal.'),
        default=None,)

    # parse arguments
    args = parser.parse_args()

    return args

###############################################################################
# FUNCTION: setup_earthdata_login_auth
###############################################################################
def setup_earthdata_login_auth(endpoint, logfile):
    """
    Set up the request library so that it authenticates against the given
    Earth data Login endpoint and is able to track cookies between requests.
    This looks in the .netrc file first and if no credentials are found,
    it prompts for them.

    Valid endpoints include:
        uat.urs.earthdata.nasa.gov - Earthdata Login UAT
        urs.earthdata.nasa.gov - Earthdata Login production

    :endpoint  - Endpoint machine name
    :logfile   - Pointer to open log file
    """
    try:
        username, _, password = netrc.netrc().authenticators(endpoint)
    except (FileNotFoundError, TypeError):
        # FileNotFound = There's no .netrc file
        # TypeError = The endpoint isn't in the netrc file, causing the
        # above to try unpacking None
        logfile.write('Please provide your Earthdata Login credentials to allow '
              'data access.\n')
        logfile.write('Your credentials will only be passed to %s and will not '
              'be exposed in Jupyter.\n' % (endpoint))
        try:
            username = input('Username:')
            password = getpass.getpass()
        except (EOFError):
            return False

    manager = request.HTTPPasswordMgrWithDefaultRealm()
    manager.add_password(None, endpoint, username, password)
    auth = request.HTTPBasicAuthHandler(manager)

    jar = CookieJar()
    processor = request.HTTPCookieProcessor(jar)
    opener = request.build_opener(auth, processor)
    request.install_opener(opener)

    return True

###############################################################################
# Function provided by Suresh Vannan at PODAAC
###############################################################################
def get_token( url: str, username: str, password: str,client_id: str, user_ip: str,logfile) -> str:
        token: str = ''
        try:
                xml: str = """<?xml version='1.0' encoding='utf-8'?>
                <token><username>{}</username><password>{}</password><client_id>{}</client_id>
                <user_ip_address>{}</user_ip_address></token>""".format(username, password, client_id, user_ip)
                headers: Dict = {'Content-Type': 'application/xml','Accept': 'application/json'}
                resp = requests.post(url, headers=headers, data=xml)
                response_content: Dict = json.loads(resp.content)
                token = response_content['token']['id']
                logfile.write("Temporary CMR token created: %s\n" % (token))
        except:
                logfile.write("Error getting the token - check user name and password.\n")
        return token

###############################################################################
# Function provided by Suresh Vannan at PODAAC
###############################################################################
def delete_token(url: str, token: str,logfile) -> None:
        try:
                headers: Dict = {'Content-Type': 'application/xml','Accept': 'application/json'}
                url = '{}/{}'.format(url, token)
                resp = requests.request('DELETE', url, headers=headers)
                if resp.status_code == 204:
                        logfile.write('CMR token successfully deleted: {}\n'.format(token))
                else:
                        logfile.write('CMR token deleting failed: {}\n'.format(token))
        except:
                logfile.write("Error getting the token - check user name and password\n")
        #exit(0)

###############################################################################
# Function: find and stack URLs from PODAAC request
###############################################################################
def find_urls(main_url, params, logfile):
    """
    Find and URLs from PODAAC request using params and main_url.
    Iterate over n_pages of retrieved results and stack URLs.

    :param main_url:    URL of main website
    :param params:      dictionary containing request parameters
    :param logfile:     log file

    :return:            list of URLs
    """
    # Define the query to get the files
    query = urllib.parse.urlencode(params)

    url_page_1 = main_url + query

    # Inital query to obtain total number of records
    # Probally a better and a cleaner way to this!
    try:
        with urllib.request.urlopen(url_page_1) as fi:
            results_page_1 = json.loads(fi.read().decode())
    except (urllib.error.HTTPError, urllib.error.URLError) as e:
        logfile.write("Error querying for ccid = %s\n" % (params))
        logfile.write("Error message:\n")
        logfile.write(str(e)+"\n")
        return []

    # Get number of total records
    n_hits = float(results_page_1['hits'])

    # Number of pages to loop trough
    n_pages = int(math.ceil((n_hits / PAGE_MAX)))

    # Get stacked page of url-link for download
    urls = []
    for page in range(1, n_pages+1):
        params['page_num'] = page
        query = urllib.parse.urlencode(params)
        url = main_url + query
        urls.append(url)

    return urls

###############################################################################
# FUNCTION: Provided CCID not available in internal dictionary: getSWOTdata
###############################################################################
def getSWOTdata(productid, downloaddir, token, logfile, twin=None, search_cycles=None, search_pass=None, meta=False, \
              bbox="-180,-90,180,90", machine=MACHINE, cmr=CMR, exts=[], patterns=[],
              updatefilepath=None, pdesc=False, listonly=False):
    """
    Function to query SWOT data from the PODACC cloud given
    specific input arguments. Provides links to the found files to download.

    :param productid:  ID of data product to download. Either has short form
                       available in this script, or an explicit ccid recognized
                       by PODAAC. Required if pdesc is False
    :param downloaddir Download data directory (Will create if does not exist)
    :param token:      Token string
    :param logfile:    Pointer to open log file
    :param twin:       Window for data search back in time (min) from now.
                       Default is 60 minutes.
                       Only used on first call. Subsequent calls look for data
                       from last download.
    :param meta:       get all files including metadata not just data.
                       Default is false.
    :param bbox:       bounding box for data 'lonmin,latmin,lonmax,latmax'.
                       Default is "-180,-90,180,90".
    :param machine:    URL of machine to download data. Default is MACHINE
    :param cmr:        CMR machine to download data. Default is CMR
    :param exts:       Only download files with specified extensions (list)
    :param patterns    Only download files matching patterns (list)
    :param updatefilepath If provided, this is the file where the last update time is saved.
                       Default uses a file name .update in the downloaddir
    :param pdesc       If True, print description of pre-defined product IDs. Default is False.
    :param listonly    If True, print list of available files within time window and do not download. 
                       Default is False with files downloaded.
    :return:           failed-list, success-list, current time in UTC.
    """

    if pdesc:
        print("Description of pre-defined product IDs. If not on list, use PODAAC ccid.")
        for x in sorted(AVAILABLE_CCIDS):
            print("%-20s : %s" % (x, CCIDDESCRIPTIONS[x]))
        sys.exit(0)
    else:
        if productid == None:
            print("Product ID id required.\nUse -h option for help, or -pdesc option for descriptions of pre-defined product IDs.")
            sys.exit(1)

    # Initialize the list of successfully and failed downloaded files
    failedlist = []
    successlist = []
    last_time = None
    filetimes = None
    # Get the current time at query
    time_now = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)

    # Check if we have a relationship between a short form ccid versus the PODAAC ccid.
    # If we do then use it. If not, then presume the user is providing the PODAAC ccid.
    if productid in CCIDDICT.keys():
        ccid = CCIDDICT[productid]
        ccidinfo = CCIDDESCRIPTIONS[productid]
        logfile.write("Started download = %s\n" % time_now.strftime("%Y-%m-%dT%H:%M:%SZ"))
        logfile.write("Short form CCID  = %s\n" % productid)
        logfile.write("PODAAC     CCID  = %s\n" % ccid)
        logfile.write("Description      = %s\n" % ccidinfo)
    else:
        ccid = productid
        logfile.write("Started download = %s\n" % time_now.strftime("%Y-%m-%dT%H:%M:%SZ"))
        logfile.write("Provided CCID not available in internal dictionary.\n")
        logfile.write("Attempting to download with assumption that it is recognized by PODAAC.\n")
        logfile.write("PODAAC     CCID = %s\n" % ccid)

    # The following steps were provided by PODAAC
    success = setup_earthdata_login_auth(machine, logfile)
    if not success:
        logfile.write("Error with setup_earthdata_login\n")
        return failedlist, successlist, time_now

    # Define the time of the last query at now for default.
    # This is the time that is used as a reference for future downloads as default...
    # but if file downloaded will use time from metadata for the youngest file downloaded

    # Define the URL to to get list of available files
    main_url = "https://" + cmr + "/search/granules.umm_json?"

    # Check for valid token
    deletetoken = False
    if token is None:
        logfile.write("A valid token was not provided. Creating one.\n")
        token = get_token(TOKEN_URL,USER_NAME,PASSWORD,'SWOT', IPAddr,logfile)
        deletetoken = True

    params = {
        'scroll': "false",
        'page_num': 1,
        'page_size': PAGE_MAX,
        'sort_key': "revision-date",
        'ShortName': ccid,
        'token': token,
        #'bounding_box': bbox,
        }

    # Get initial query to get number of hits
    if search_cycles is not None:
        # Iterate over Cycles
        cmr_cycles = search_cycles
        urls = []
        for v in cmr_cycles:
            logfile.write(f"Adding URLs for cycle {v}\n")
            params["cycle[]"] = v
            # In case both cycle and pass are provided, retrieve result directly via URL
            if search_pass is not None:
                logfile.write(f"... and pass {search_pass}\n")
                params['passes[0][pass]'] = search_pass
            urls_ = find_urls(main_url, params, logfile)
            urls += urls_
    else:
        if twin is not None:
            # Check time window for overflow
            if twin > MINMAX:
                twin = MINMAX
                logfile.write(f"Time window to large! Setting to max value {MINMAX}!.\n")
            # Define the search window
            tsearch = (time_now - datetime.timedelta(minutes=twin)).strftime("%Y-%m-%dT%H:%M:%SZ")
            logfile.write("Seach window starts at: %s\n" % (tsearch))
            params['created_at'] = tsearch
        # If no filter criteria are provided, retrieve all available results
        urls = find_urls(main_url, params, logfile)
    

    # Initialize list of files to download
    links = []
    for url in urls:
        # Perform the query for available files
        try:
            with urllib.request.urlopen(url) as f:
                results = json.loads(f.read().decode())
        except (urllib.error.HTTPError, urllib.error.URLError) as e:
            logfile.write("Error querying for ccid = %s\n" % (ccid))
            logfile.write("Error message:\n")
            logfile.write(str(e)+"\n")
            return failedlist, successlist, time_now

        # List of raw data files
        for r in results['items']:
            revision_time = datetime.datetime.strptime(
                    r['meta']['revision-date'],
                    "%Y-%m-%dT%H:%M:%S.%fZ"
                ).replace(tzinfo=datetime.timezone.utc)
            deltatime = revision_time - time_now
            for u in r['umm']['RelatedUrls']:
                if u['Type'] == "GET DATA" and ('Subtype' not in u or u['Subtype'] != "OPENDAP DATA"):
                    link = u['URL']
                    base = basename(link)
                    saveext = False
                    savepatt = False
                    if len(exts) != 0:
                        for ext in exts:
                            if base.endswith(ext):
                                saveext = True
                    else:
                        saveext = True
                    if len(patterns) != 0:
                        for pattern in patterns:
                            if pattern in base:
                                savepatt = True
                    else:
                        savepatt = True
                    # collect links only if they match extension and pattern
                    if saveext and savepatt:
                        links.append([link, revision_time, deltatime])

    # Number of files
    logfile.write("%d new files found for download.\n" % (len(links)))

    # Sort files by delta time so oldest file downloaded first
    sortedlinks = sorted(links, key=lambda link: link[2])

    # Print list of files found
    if listonly:
        print("\nNot downloading files.")
        print("Available files within specificed time window as follows: file, revistion date, age from now.")
        for link, rev_date, deltatime in sortedlinks:
            basefile = basename(link)
            print(basefile, rev_date, deltatime)
        sys.exit(0)

    # Check for valid download directory
    if downloaddir is None:
        downloaddir = ccid
        logfile.write(f"Download directory, not provided. Using CCID as default: {ccid}\n")

    # Create the download directory if necessary
    if not os.path.exists(downloaddir):
        try:
            os.makedirs(downloaddir)
        except OSError:
            logfile.write("Unable to create directory: %s\n" % (downloaddir))
            return failedlist, successlist, time_now

    # Define the update file
    if updatefilepath is None:
        updatefile = os.path.join(downloaddir, UPDATEFILE)
    else:
        updatefile = updatefilepath

    # Read the update file
    if updatefile is not None and os.path.isfile(updatefile):
        # Read the header time which contains last update time
        # Read the individual lines for prior files downloaded within window and their
        # revision times as defined by PODAAC
        with open(updatefile, "r") as f:
            filetimes = yaml.safe_load(f)
        last_time = filetimes.pop('Last_query') # pop returns value and removes item from dict
        # Generate list of files previously downloaded
        prevfiles = [k for k in filetimes.keys() if k != "Last_query"]
        logfile.write("Prior update file found %s. Last update at time %s. %d files stored.\n" % (updatefile, last_time, len(prevfiles)))
    else:
        prevfiles = []
        logfile.write("First download. No previous update file (%s) found.\n" % (updatefile))

    # Override tlastquery to the time at which the last query was made
    #tlastquery = time_now
    if updatefile is not None:
        with open(updatefile, "w") as f:
            # Write header containing last query time (only for info)
            f.write("Last_query: %s\n" % datetime.datetime.strftime(time_now,"%Y-%m-%dT%H:%M:%S.%fZ"))
            if filetimes is not None:
                for fn, t in filetimes.items():
                    f.write("%s: %s\n" % (fn, datetime.datetime.strftime(t,"%Y-%m-%dT%H:%M:%S.%fZ")))


    # Loop through each link and try to download
    skippedlist = []
    for link, revdate, _ in sortedlinks:
        # Check if file already downloaded, and skip if found file has revision date
        # the same or earlier than previously found file
        basefile = basename(link)
        if (basefile in prevfiles) and (revdate <= filetimes[basefile].replace(tzinfo=datetime.timezone.utc)):
            logfile.write(f"Skipping {basefile}\n")
            skippedlist.append(basefile)
            continue
        try:
            urllib.request.urlretrieve(link, os.path.join(downloaddir, basefile))
        except (urllib.error.HTTPError, urllib.error.URLError) as e:
            logfile.write("Error with urlretrieve.\n")
            logfile.write("Error message:\n")
            logfile.write(str(e)+"\n")
            failedlist.append(basefile)
        else:
            successlist.append(basefile)
            # append downloaded filenames and revision date to update file
            if os.path.exists(updatefile):
                with open(updatefile, "a") as f:
                    tdelta = (time_now - revdate).total_seconds()/60
                    if search_cycles is not None:
                        f.write("%s: %s\n" % (basefile, datetime.datetime.strftime(revdate,"%Y-%m-%dT%H:%M:%S.%fZ")))
                    elif twin is not None:
                        if tdelta <= twin:
                            f.write("%s: %s\n" % (basefile, datetime.datetime.strftime(revdate,"%Y-%m-%dT%H:%M:%S.%fZ")))


    logfile.write("Successfully downloaded files: %d\n" % (len(successlist)))
    logfile.write("Failed not   downloaded files: %d\n" % (len(failedlist)))
    logfile.write("Previously   downloaded files: %d\n" % (len(skippedlist)))

    # Delete the token if it was created on the fly
    if deletetoken:
        delete_token(TOKEN_URL,token,logfile)

    return failedlist, successlist, time_now


def main():
    # Read input arguments
    args = parseArgs()

    # Get parameters
    tspan    = args.twin
    ccid     = args.prodid
    ddir     = args.ddir
    token    = args.token
    sleep    = args.sleep
    extlist  = args.extlist
    pattlist = args.pattlist
    logfile  = args.logfile
    search_cycles = args.search_cycles
    search_pass = args.search_pass
    log_file = sys.__stdout__

    f_fail, f_ok, time_last = getSWOTdata(ccid, ddir, token, log_file, \
                                  tspan, search_cycles, search_pass, exts=extlist, patterns=pattlist)


if __name__ == '__main__':
    main()
